/*
SPLICE 1.2 : a C++ utility for "gene" splicing, cutting, copying, inserting and
        overlaying bits of a file into another
        WARNING SPLICE ONLY HANDLES 8.3 FILENAMES.

        Uses for SPLICE:
        1. excising a section of a file.
        2. replacing a section of a file.
        3. inserting a section into a file.
        4. extracting a section of a file.
        5. appending a section onto a file.

        Splice 1.2
        This program may be copied freely and used for any purpose except military.
        */

/*
Roedy Green
Canadian Mind Products
#208 - 525 Ninth Street
New Westminster BC Canada
V5H 2N6
tel:(604) 777-1804
mailto:roedy@mindprod.com
http://mindprod.com
*/

/*
 compiled under SMALL model for DOS under Borland C++ 5.01

SPLICE commands:

CUT MyFile1.Txt START 6 LENGTH 5 RECSIZE 1024
    remove a slice from a file offset 6 length 5
    Offsets and lengths are measured in blocks of 1024.

CUT MyFile1.Txt START 6 LENGTH 5 TO MyFile2.txt RECSIZE 1024
    remove a slice from a file offset 6 length 5
    and copy it to a new file.

CUT MyFile1.Txt START 6 LENGTH 5 APPEND MyFile2.txt RECSIZE 1024
    remove a slice from a file offset 6 length 5
    and copy it to the end of another file.

CUT MyFile1.Txt START 6 LENGTH 5 OVERLAY MyFile2.txt AT 7 RECSIZE 1024
    remove a slice from a file, offset 6 length 5,
    overlay it on top of another file at offset 7.

CUT MyFile1.Txt START 6 LENGTH 5 INSERT MyFile2.txt AT 7 RECSIZE 1024
    remove a slice from a file, offset 6, length 5, insert
    at offset 7 in another file

COPY MyFile1.Txt START 6 LENGTH 5 TO MyFile2.txt RECSIZE 1024
    copy a slice from a file offset 6 length 5
    to a new file

COPY MyFile1.Txt START 6 LENGTH 5 APPEND MyFile2.txt RECSIZE 1024
    copy a slice from a file offset 6 length 5
    to the end of another file.

COPY MyFile1.Txt START 6 LENGTH 5 OVERLAY MyFile2.txt AT 7 RECSIZE 1024
    copy a slice from a file, offset 6 length 5,
    overlaying it on top of another file at offset 7,
    replacing the previous data.

COPY MyFile1.Txt START 6 LENGTH 5 INSERT MyFile2.txt AT 7 RECSIZE 1024
    copy a slice from a file, offset 6, length 5, inserting
    at offset 7 in another file.

START defaults to 0
AT defaults to 0
RECSIZE defaults to 1
If length reaches out past the end of the source file, it will be shortened.
Target files will be expanded as needed with zeroes.

Version 1.2 1998 November 8
        - embed Barker address
        - correct Bulk::move logic so that fromOffset is defined.

Version 1.1 1997 April 28
        - added APPEND
        - add abbreviations for keywords

Version 1.0 1997 April 27
        - initial release


Future consideration:
        - long filename support.

*/

/* I N C L U D E S */

#include <stdlib.h>
// e.g. exit, min, max

#include <stdio.h>
// e.g. FILE, printf, fopen

#include <fcntl.h>
// O_CREAT O_EXCL
#include <sys\stat.h>
// S_IWRITE

#include <io.h>
// open, write

#include <conio.h>
// e.g. getch kbhit crlscr

#include <string.h>
// strcpy

/* P R O T O T Y P E S */

int main( int argc,  char *argv[] );
void oops(char* reason);
void banner();

// =============================================================================

class SeqFile
  {

public:
  // size of file in bytes
  long size;
  int handle;

  // we use DOS level handle open rather than fopen to get access to a way
  // to shrink files.

  // constructor
  SeqFile (char* internalName, char* fileName, int accessFlags, int createFlags);

  // destructor
  ~SeqFile();

  }; // end SeqFile

// =============================================================================

class Bulk
  {
  // bulk moving data around in and between files


// buffer size is 31K.  This makes it a 16-bit signed value.
#define bufferSize 31744

public:

  // clear area to constant
  static void zap(int theFile, long start, long length, unsigned char value);

  // move data within a file, either up or down
  static void move(int theFile, long start, long length, long at);

  // move data between two different files
  static void move(int fromFile, long start, long length, int toFile, long at);

  static void resize(int theFile, long size);

  }; // end Bulk

// =============================================================================

typedef enum SourceType
  {
  unknown, cut, copy
  } SourceType;

typedef enum TargetType
  {
  none, to, append, overlay, insert
  } TargetType;

class Splice
  {

public:

  static void splice(int argc,  char *argv[]);

protected:

  // files
  static SeqFile* sourceFile;
  static SeqFile* targetFile;

  // offset control
  static long  start;
  static long  length;
  static long  at;
  static long  recsize;

  // source/target kind
  static SourceType source;
  static TargetType target;

  static void parse( int argc,  char *argv[] );
  static void checkParms();
  static void openFiles();
  static void analyse();
  static void work();
  static void closeFiles();

  // parms from the command line
  static char* cutName;
  static char* copyName;
  static char* toName;
  static char* appendName;
  static char* overlayName;
  static char* insertName;

  }; // end class Splice
// =============================================================================

/* S T A T I C   D E F I N I T I O N S   A N D   I N I T I A L I S A T I O N */

// offset control
long  Splice::start    = 0;
long  Splice::length   = 0;
long  Splice::at       = 0;
long  Splice::recsize  = 1;

// source/target
SourceType Splice::source = unknown;
TargetType Splice::target = none;

SeqFile* Splice::sourceFile = NULL;
SeqFile* Splice::targetFile = NULL;

// parms from the command line
char* Splice::cutName       = NULL;
char* Splice::copyName      = NULL;
char* Splice::toName        = NULL;
char* Splice::appendName    = NULL;
char* Splice::overlayName   = NULL;
char* Splice::insertName    = NULL;

// =============================================================================

/* M E T H O D   D E F I N I T I O N S */

SeqFile::SeqFile (char* internalName, char* fileName, int accessFlags, int createFlags)

{ // constructor

  size = 0;
  handle = -1;
  if ( fileName && ( handle = open( fileName, accessFlags, createFlags ) ) == -1 )
    {
    char* complaint = new char[256];
    complaint[0] = 0;
    strcat(complaint, "Can't open ");
    strcat(complaint, internalName);
    strcat(complaint, " file ");
    strcat(complaint, fileName);
    strcat(complaint, ".");
    oops(complaint);  // won't return
    delete complaint;
    }
  if ( handle != -1 ) size = filelength(handle);

} // end SeqFile constructor

SeqFile::~SeqFile ()
{
  // destructor
  if ( handle ) close(handle);
} // end SeqFile destructor

// =============================================================================

void Bulk::zap(int theFile, long start, long length, unsigned char value)
{
  // clear area of file to a value
  if ( length == 0 ) return;

  unsigned char* buffer = new unsigned char [bufferSize];
  if ( buffer == NULL ) oops ("Not enough RAM.");
  for ( int i=0; i<bufferSize; i++ )
    {
    buffer[i] = value;
    }

  lseek (theFile, start, SEEK_SET);

  // do big chunks first
  long end = start + ((length / bufferSize) * bufferSize);
  for ( long offset=start; offset<end; offset+=bufferSize )
    {
    write(theFile, buffer, bufferSize);
    } // end for

  // do the remainder
  size_t remainder = (size_t)(length % bufferSize);
  if ( remainder )
    {
    write(theFile, buffer, remainder);
    }

  delete buffer;

} // end zap

void Bulk::move(int theFile, long start, long length, long at)
{
  // move data within a file, either up or down

  if ( start == at || length == 0 ) return;

  unsigned char* buffer = new unsigned char [bufferSize];
  if ( buffer == NULL ) oops ("Not enough RAM.");

  if ( start < at && at < start + length )
    {
    // because of overlap, work down, right to left
    long begin = start + length - bufferSize;
    long end = begin - ((length / bufferSize) * bufferSize);
    long toOffset = at + length - bufferSize;
    long fromOffset;
    for ( fromOffset=begin; fromOffset>end; fromOffset-=bufferSize )
      {
      lseek (theFile, fromOffset, SEEK_SET);
      read (theFile, buffer, bufferSize);
      lseek (theFile, toOffset, SEEK_SET);
      write(theFile, buffer, bufferSize);
      toOffset -= bufferSize;
      }

    // do the remainder
    size_t remainder = (size_t) (length % bufferSize);
    if ( remainder )
      {
      lseek (theFile, start, SEEK_SET);
      read (theFile, buffer, remainder);
      lseek (theFile, at, SEEK_SET);
      write(theFile, buffer, remainder);
      }
    } else
    {
    // work up, left to right.
    // end is offset just past last big chunk
    long end = start + ((length / bufferSize) * bufferSize);
    long toOffset = at;
    long fromOffset;
    for ( fromOffset=start; fromOffset<end; fromOffset+=bufferSize )
      {
      lseek (theFile, fromOffset, SEEK_SET);
      read (theFile, buffer, bufferSize);
      lseek (theFile, toOffset, SEEK_SET);
      write(theFile, buffer, bufferSize);
      toOffset += bufferSize;
      }

    // do the remainder
    size_t remainder = (size_t)(length % bufferSize);
    if ( remainder )
      {
      lseek (theFile, fromOffset, SEEK_SET);
      read (theFile, buffer, remainder);
      lseek (theFile, toOffset, SEEK_SET);
      write(theFile, buffer, remainder);
      }
    }
  delete buffer;
} // end movedata

void Bulk::move(int fromFile, long start, long length, int toFile, long at)
{
  // move data between two different files
  if ( length == 0 ) return;

  unsigned char* buffer = new unsigned char [bufferSize];
  if ( buffer == NULL ) oops ("Not enough RAM.");

  lseek (fromFile, start, SEEK_SET);
  lseek (toFile, at, SEEK_SET);

  // do big chunks first
  long end = start + ((length / bufferSize) * bufferSize);
  for ( long offset=start; offset<end; offset+=bufferSize )
    {
    read (fromFile, buffer, bufferSize);
    write(toFile, buffer, bufferSize);
    } // end for

  // do the remainder
  size_t remainder = (size_t) (length % bufferSize);
  if ( remainder )
    {
    read (fromFile, buffer, remainder);
    write(toFile, buffer, remainder);
    }

  delete buffer;
} // end movedata

void Bulk::resize (int theFile, long size)
{

  // defeat the auto zero-padding feature on growth
  // since we will be filling the growth area ourselves.
  long oldsize = filelength(theFile);
  if ( oldsize + 1 < size )
    {
    lseek(theFile,size-1,SEEK_SET);
    write(theFile,0,1); // write one null byte to grow the file
    }
  // may be shrinking or growing
  if ( chsize(theFile, size) == -1 ) oops ("insufficient disk space");

} // end resize

// =============================================================================

void Splice::checkParms()
{
  char * sourceName = NULL;
  source = unknown;
  int counter = 0;
  if ( cutName )
    {
    source = cut;  sourceName = cutName;  counter++;
    }
  if ( copyName )
    {
    source = copy; sourceName = copyName; counter++;
    }
  if ( counter != 1 )
    oops("You must have either a CUT or COPY source.");

  char * targetName = NULL;
  target = none;
  counter = 0;
  if ( toName )
    {
    target = to;      targetName = toName;      counter++;
    }
  if ( appendName )
    {
    target = append;  targetName = appendName;  counter++;
    }
  if ( overlayName )
    {
    target = overlay; targetName = overlayName; counter++;
    }
  if ( insertName )
    {
    target = insert;  targetName = insertName;  counter++;
    }
  if ( !(counter == 1  || (counter == 0 && source == cut)) )
    oops ("You must have either a TO, OVERLAY or INSERT target.");

  if ( stricmp(sourceName, targetName) == 0 )
    oops ("You can't have the source and target files the same.");

  if ( target == to && at )
    oops ("TO and AT are incompatible. Use OVERLAY or INSERT");

  if ( target == append && at )
    oops ("APPEND and AT are incompatible. Use OVERLAY or INSERT");

  if ( source == cut && target == none && at )
    oops ("Use START, not AT to CUT");

  if ( start < 0 )
    oops ("START must be >= 0");

  if ( length  <= 0 )
    oops ("LENGTH must be > 0");

  if ( at < 0 )
    oops ("AT must be >= 0");

  if ( recsize <= 0 )
    oops ("RECSIZE must be > 0");

} // end checkParms

void Splice::parse ( int argc,  char *argv[] )
{
  // parse the command line
  char*  keyword;
  char*  value;
  for ( int i=1; i<argc; i+=2 )
    {
    keyword = argv[i];
    if ( i+1 >= argc ) oops("Must have keyword/value pairs separarated by space");
    value = argv[i+1];

    if ( stricmp(keyword,"append")      ==0 ) appendName = value;
    else if ( stricmp(keyword,"app")    ==0 ) appendName = value;

    else if ( stricmp(keyword,"at")     ==0 ) at        = atol(value);

    else if ( stricmp(keyword,"copy")   ==0 ) copyName  = value;

    else if ( stricmp(keyword,"cut")    ==0 ) cutName   = value;

    else if ( stricmp(keyword,"insert") ==0 ) insertName  = value;
    else if ( stricmp(keyword,"ins")    ==0 ) insertName  = value;
    else if ( stricmp(keyword,"i")      ==0 ) insertName  = value;

    else if ( stricmp(keyword,"length") ==0 ) length    = atol(value);
    else if ( stricmp(keyword,"len")    ==0 ) length    = atol(value);
    else if ( stricmp(keyword,"l")      ==0 ) length    = atol(value);

    else if ( stricmp(keyword,"overlay")==0 ) overlayName = value;
    else if ( stricmp(keyword,"over")   ==0 ) overlayName = value;
    else if ( stricmp(keyword,"o")      ==0 ) overlayName = value;

    else if ( stricmp(keyword,"recsize")==0 ) recsize   = atol(value);
    else if ( stricmp(keyword,"lrecl")  ==0 ) recsize   = atol(value);
    else if ( stricmp(keyword,"by")     ==0 ) recsize   = atol(value);
    else if ( stricmp(keyword,"r")      ==0 ) recsize   = atol(value);

    else if ( stricmp(keyword,"start")  ==0 ) start     = atol(value);
    else if ( stricmp(keyword,"s")      ==0 ) start     = atol(value);

    else if ( stricmp(keyword,"to")     ==0 ) toName    = value;
    else if ( stricmp(keyword,"t")      ==0 ) toName    = value;

    else oops("unrecogisable or missing keyword");
    }
} // end parse

void Splice::openFiles()
{

  // open source and target files
  switch ( source )
    {
    case cut:
      sourceFile  = new SeqFile("CUT",cutName,O_RDWR+O_BINARY,0);
      break;
    case copy:
      sourceFile  = new SeqFile("COPY",copyName,O_RDONLY+O_BINARY,0);
      break;
    case unknown:
    default:
      oops("bug #1 in program");
    } // end switch

  switch ( target )
    {
    case none:
      targetFile = NULL;
      break;
    case to:
      targetFile  = new SeqFile("TO",toName,
                                O_WRONLY+O_BINARY+O_TRUNC+O_CREAT,
                                S_IREAD+S_IWRITE);
      break;
    case append:
      targetFile = new SeqFile("APPEND",appendName,O_WRONLY+O_BINARY,0);
      break;
    case overlay:
      targetFile  = new SeqFile("OVERLAY",overlayName,O_WRONLY+O_BINARY,0);
      break;
    case insert:
      targetFile  = new SeqFile("INSERT",insertName,O_RDWR+O_BINARY,0);
      break;
    default:
      oops("bug #2 in program.");
    } // end switch
} // end openFiles

void Splice::analyse ()
{
  start = start * recsize;
  length = length * recsize;
  at = at * recsize;

  // treat append like OVERLAY AT
  if ( target == append ) at = targetFile->size;

  // reign in regions that slop outside the sourceFile
  if ( start >= sourceFile->size ) length = 0;
  else if ( start+length > sourceFile->size ) length = sourceFile->size - start;

  // regions on the target may well slop outside TargetFile
  // so we don't reign them in. We pad with zeroes.

} // end analyse

void Splice::work()
{
  if ( target != none )
    {
    // calculate how big the target file is going to be
    // so we can allocate its space in one fell swoop.
    long newTargetSize = targetFile->size;
    if ( at > newTargetSize ) newTargetSize = at;
    if ( target == insert ) newTargetSize += length;
    if ( at + length > newTargetSize ) newTargetSize = at + length;
    if ( newTargetSize != targetFile->size )
      Bulk::resize(targetFile->handle, newTargetSize);
    }

  // pad with zeroes if we are overlaying or inserting past the end
  if ( target != none && (at > targetFile->size) )
    {

    Bulk::zap(targetFile->handle,
              targetFile->size, // old size
              at - targetFile->size,
              0);
    }

  // open up a hole in the target if needed.
  if ( target == insert )
    {
    Bulk::move(targetFile->handle,
               at,
               targetFile->size /* old size */ - at,
               at + length);
    }

  // copy the contents of the region to the second file if needed.
  if ( target != none )
    {
    Bulk::move(sourceFile->handle,
               start,
               length,
               targetFile->handle,
               at);
    }

  // close up the hole in the source if needed.
  if ( source == cut )
    {
    // move data down onto the hole
    Bulk::move(sourceFile->handle,
               start + length,
               sourceFile->size - (start + length),
               start);
    // shrink source file
    Bulk::resize(sourceFile->handle,
                 sourceFile->size - length);
    }

} // end work

void Splice::closeFiles()
{
  if ( sourceFile ) delete sourceFile;
  if ( targetFile ) delete targetFile;

} // end closeFiles

void Splice::splice(int argc,  char *argv[])
{
  // master routine to handle the splice command line
  parse(argc,argv);
  checkParms();
  openFiles();
  analyse();
  work();
  closeFiles();
}
// =============================================================================

void oops(char* reason)
{
  banner();
  printf("Oops, error in SPLICE command line\n");
  printf(reason);
  printf("\ntry:\n"
         "SPLICE CUT  MyFile1.Txt START 99 LENGTH 99                           RECSIZE 1\n"
         "SPLICE CUT  MyFile1.Txt START 99 LENGTH 99 TO      MyFile2.txt       RECSIZE 1\n"
         "SPLICE CUT  MyFile1.Txt START 99 LENGTH 99 APPEND  MyFile2.txt       RECSIZE 1\n"
         "SPLICE CUT  MyFile1.Txt START 99 LENGTH 99 OVERLAY MyFile2.txt AT 99 RECSIZE 1\n"
         "SPLICE CUT  MyFile1.Txt START 99 LENGTH 99 INSERT  MyFile2.txt AT 99 RECSIZE 1\n"
         "SPLICE COPY MyFile1.Txt START 99 LENGTH 99 TO      MyFile2.txt       RECSIZE 1\n"
         "SPLICE COPY MyFile1.Txt START 99 LENGTH 99 APPEND  MyFile2.txt       RECSIZE 1\n"
         "SPLICE COPY MyFile1.Txt START 99 LENGTH 99 OVERLAY MyFile2.txt AT 99 RECSIZE 1\n"
         "SPLICE COPY MyFile1.Txt START 99 LENGTH 99 INSERT  MyFile2.txt AT 99 RECSIZE 1\n");

  exit (2);
} // end oops


void banner(void)
{
  /* display copyright banner. */

  clrscr(); /* clear screen */
  printf("\n" "\xb0\xb1\xb2\xdb" " Splice 1.2 "  "\xdb\xb2\xb1\xb0"
         "\n"
         "\nSplices a chunk of a file into another file."
         "\n"
         "\nCopyright (c) 1997,1998 Canadian Mind Products"
         "\n" "#208 - 525 Ninth Street, New Westminster, BC Canada V3M 5T9"
         "\n" "tel:(604) 777-1804   mailto:roedy@mindprod.com   http://mindprod.com"
         "\n" "May be used freely for non-military use only"
         "\n\n");

} /* end Banner */

/*****************/

int main( int argc,  char *argv[] )
{
  Splice::splice(argc,argv);
  return 0;
} // end main
// =============================================================================
